"""Utility functions to be used in other scripts."""

import gc
import importlib
import logging
import os
import random
import re
import warnings
from typing import Optional, Sequence, Tuple, Type, Union

import numpy as np
import pkg_resources
import requests
import torch
from datasets.utils import disable_progress_bar
from requests.exceptions import RequestException
from transformers import logging as tf_logging
from transformers.tokenization_utils import PreTrainedTokenizer

from .exceptions import InvalidBenchmark


def clear_memory():
    """Clears the memory of unused items."""

    # Clear the Python cache
    gc.collect()

    # Empty the CUDA cache
    # TODO: Also empty MPS cache
    if torch.cuda.is_available():
        torch.cuda.empty_cache()


def enforce_reproducibility(framework: str, seed: int = 4242):
    """Ensures reproducibility of experiments.

    Args:
        framework (str):
            The framework used for the benchmarking.
        seed (int):
            Seed for the random number generator.
    """
    random.seed(seed)
    np.random.seed(seed)
    rng = np.random.default_rng(seed)
    if framework in ("pytorch", "jax"):
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.use_deterministic_algorithms(True, warn_only=True)
    return rng


def is_module_installed(module: str) -> bool:
    """Check if a module is installed.

    This is used when dealing with spaCy models, as these are installed as separate
    Python packages.

    Args:
        module (str):
            The name of the module.

    Returns:
        bool:
            Whether the module is installed or not.
    """
    # Get list of all modules, including their versions
    installed_modules_with_versions = list(pkg_resources.working_set)

    # Strip the module versions from the list of modules. Also make the modules lower
    # case and replace dashes with underscores
    installed_modules = [
        re.sub("[0-9. ]", "", str(module)).lower().replace("-", "_")
        for module in installed_modules_with_versions
    ]

    # Check if the module is installed by checking if the module name is in the list
    return module.lower() in installed_modules


def block_terminal_output():
    """Blocks libraries from writing output to the terminal.

    This filters warnings from some libraries, sets the logging level to ERROR for some
    libraries, disabled tokeniser progress bars when using Hugging Face tokenisers, and
    disables most of the logging from the `transformers` library.
    """

    # Ignore miscellaneous warnings
    warnings.filterwarnings(
        "ignore",
        module="torch.nn.parallel*",
        message="Was asked to gather along dimension 0, but all input tensors were "
        "scalars; will instead unsqueeze and return a vector.",
    )
    warnings.filterwarnings("ignore", module="seqeval*")

    # Up the logging level, to disable outputs
    logging.getLogger("filelock").setLevel(logging.ERROR)
    logging.getLogger("absl").setLevel(logging.ERROR)
    logging.getLogger("datasets").setLevel(logging.ERROR)

    # Disable the tokeniser progress bars
    disable_progress_bar()

    # Disable most of the `transformers` logging
    tf_logging._default_log_level = logging.CRITICAL
    tf_logging.set_verbosity(logging.CRITICAL)
    logging.getLogger("transformers.trainer").setLevel(logging.CRITICAL)


def get_class_by_name(
    class_name: Union[str, Sequence[str]],
    module_name: Optional[str] = None,
) -> Union[None, Type]:
    """Get a class by its name.

    Args:
        class_name (str or list of str):
            The name of the class, written in kebab-case. The corresponding class name
            must be the same, but written in PascalCase, and lying in a module with the
            same name, but written in snake_case. If a list of strings is passed, the
            first class that is found is returned.
        module_name (str, optional):
            The name of the module where the class is located. If None then the module
            name is assumed to be the same as the class name, but written in
            snake_case. Defaults to None.

    Returns:
        type or None:
            The class. If the class is not found, None is returned.
    """
    # Ensure that `class_name` is a sequence
    if isinstance(class_name, str):
        class_name = [class_name]

    # Loop over the class names
    for name in class_name:

        # Get the snake_case and PascalCase version of the class name
        name_snake = name.replace("-", "_")
        name_pascal = kebab_to_pascal(name)

        # Import the module
        try:
            if not module_name:
                module_name = f"scandeval.{name_snake}"
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            module_name = None
            continue

        # Get the class from the module
        try:
            class_: Type = getattr(module, name_pascal)
        except AttributeError:
            module_name = None
            continue

        # Return the class
        return class_

    # If the class could not be found, return None
    return None


def kebab_to_pascal(kebab_string: str) -> str:
    """Converts a kebab-case string to PascalCase.

    Args:
        kebab_string (str):
            The kebab-case string.

    Returns:
        str:
            The PascalCase string.
    """
    return "".join(word.title() for word in kebab_string.split("-"))


def handle_error(
    e: Exception, per_device_train_batch_size: int, gradient_accumulation_steps: int
) -> Tuple[int, int]:
    """Handle an error that occurred during the benchmarking process.

    Args:
        e (Exception):
            The exception that was raised.
        per_device_train_batch_size (int):
            The batch size used for training.
        gradient_accumulation_steps (int):
            The number of gradient accumulation steps.

    Returns:
        pair of int:
            The batch size and gradient accumulation steps to use.
    """
    # We assume that all these CUDA errors are caused by
    # insufficient GPU memory
    # TODO: Handle MPS out of memory as well
    cuda_errs = ["CUDA out of memory", "CUDA error"]

    # If it is an unknown error, then simply report it
    if all([err not in str(e) for err in cuda_errs]):
        raise InvalidBenchmark(str(e))

    # If it is a CUDA memory error, then reduce batch size and up gradient
    # accumulation
    if per_device_train_batch_size == 1:
        raise InvalidBenchmark("CUDA out of memory, even with a batch size of 1!")
    return per_device_train_batch_size // 2, gradient_accumulation_steps * 2


def internet_connection_available() -> bool:
    """Checks if internet connection is available by pinging google.com.

    Returns:
            bool:
                Whether or not internet connection is available.
    """
    try:
        requests.get("https://www.google.com")
        return True
    except RequestException:
        return False


def get_special_token_metadata(tokenizer: PreTrainedTokenizer) -> dict:
    """Get the special token metadata for a tokenizer.

    Args:
        tokenizer (PreTrainedTokenizer):
            The tokenizer.

    Returns:
        dict:
            The special token metadata.
    """
    # Create some test input IDs, to check if the tokenizer is adding special tokens
    test_input_ids = tokenizer("Test").input_ids

    # Extract the CLS token IDs from the tokenizer, if it's using them
    has_cls_token = True
    if tokenizer.cls_token_id in test_input_ids:
        cls_token_id = tokenizer.cls_token_id
        cls_token = tokenizer.cls_token
    elif tokenizer.bos_token_id in test_input_ids:
        cls_token_id = tokenizer.bos_token_id
        cls_token = tokenizer.bos_token
    elif tokenizer.cls_token is not None:
        cls_token_id = tokenizer.cls_token_id
        cls_token = tokenizer.cls_token
        has_cls_token = False
    else:
        cls_token_id = tokenizer.bos_token_id
        cls_token = tokenizer.bos_token
        has_cls_token = False

    # Extract the SEP token IDs from the tokenizer, if it's using them
    has_sep_token = True
    if tokenizer.sep_token_id in test_input_ids:
        sep_token = tokenizer.sep_token
    elif tokenizer.eos_token_id in test_input_ids:
        sep_token = tokenizer.eos_token
    elif tokenizer.sep_token is not None:
        sep_token = tokenizer.sep_token
        has_sep_token = False
    else:
        sep_token = tokenizer.eos_token
        has_sep_token = False

    return dict(
        cls_token_id=cls_token_id,
        cls_token=cls_token,
        sep_token=sep_token,
        has_cls_token=has_cls_token,
        has_sep_token=has_sep_token,
    )



"""Functions related to fetching data from the Hugging Face Hub."""

import logging
import re
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Optional, Sequence, Union

from huggingface_hub.hf_api import HfApi, ModelInfo
from huggingface_hub.utils.endpoint_helpers import ModelFilter
from requests.exceptions import RequestException

from .config import BenchmarkConfig, Language, ModelConfig
from .exceptions import HuggingFaceHubDown, InvalidBenchmark, NoInternetConnection
from .languages import DA, NB, NN, NO, SV, get_all_languages
from .utils import internet_connection_available

logger = logging.getLogger(__name__)


def get_model_config(model_id: str, benchmark_config: BenchmarkConfig) -> ModelConfig:
    """Fetches configuration for a model from the Hugging Face Hub.

    Args:
        model_id (str):
            The full Hugging Face Hub ID of the model.
        benchmark_config (BenchmarkConfig):
            The configuration of the benchmark.

    Returns:
        ModelConfig:
            The model configuration.

    Raises:
        RuntimeError:
            If the extracted framework is not recognized.
    """
    # If the model ID specifies a fresh ID, then return a hardcoded metadata
    # dictionary
    if model_id.startswith("fresh"):
        model_config = ModelConfig(
            model_id=model_id,
            framework="pytorch",
            task="fill-mask",
            languages=list(),
            revision="main",
        )
        return model_config

    # Extract the revision from the model ID, if it is specified
    if "@" in model_id:
        model_id_without_revision, revision = model_id.split("@", 1)
    else:
        model_id_without_revision = model_id
        revision = "main"

    # Extract the author and model name from the model ID
    author: Optional[str]
    if "/" in model_id_without_revision:
        author, model_name = model_id_without_revision.split("/")
    else:
        author = None
        model_name = model_id_without_revision

    # Attempt to fetch model data from the Hugging Face Hub
    try:

        # Define the API object
        api: HfApi = HfApi()

        # Fetch the model metadata
        models = api.list_models(
            filter=ModelFilter(author=author, model_name=model_name),
            use_auth_token=benchmark_config.use_auth_token,
        )

        # Filter the models to only keep the one with the specified model ID
        models = [
            model for model in models if model.modelId == model_id_without_revision
        ]

        # Check that the model exists. If it does not then raise an error
        if len(models) == 0:
            raise InvalidBenchmark(
                f"The model {model_id} does not exist on the Hugging Face Hub."
            )

        # Fetch the model tags
        tags: Sequence[str] = models[0].tags

        # Extract the framework, which defaults to PyTorch
        framework = "pytorch"
        if "pytorch" in tags:
            pass
        elif "jax" in tags:
            framework = "jax"
        elif "spacy" in tags:
            raise InvalidBenchmark("SpaCy models are not supported.")
        elif "tf" in tags or "tensorflow" in tags or "keras" in tags:
            raise InvalidBenchmark("TensorFlow/Keras models are not supported.")

        # Extract the model task, which defaults to 'fill-mask'
        model_task: Optional[str] = models[0].pipeline_tag
        if model_task is None:
            model_task = "fill-mask"

        # Get list of all language codes
        language_mapping = get_all_languages()
        language_codes = list(language_mapping.keys())

        # Construct the model config
        model_config = ModelConfig(
            model_id=models[0].modelId,
            framework=framework,
            task=model_task,
            languages=[language_mapping[tag] for tag in tags if tag in language_codes],
            revision=revision,
        )

    # If fetching from the Hugging Face Hub failed then throw a reasonable exception
    except RequestException:
        if internet_connection_available():
            raise HuggingFaceHubDown()
        else:
            raise NoInternetConnection()

    # Return the model config
    return model_config


# TODO: Cache this
def get_model_lists(
    languages: Optional[Sequence[Language]],
    use_auth_token: Union[bool, str],
) -> Dict[str, Sequence[str]]:
    """Fetches up-to-date model lists.

    Args:
        languages (None or sequence of Language objects):
            The language codes of the language to consider. If None then the models
            will not be filtered on language.
        use_auth_token (bool or str):
            The authentication token for the Hugging Face Hub. If a boolean value is
            specified then the token will be fetched from the Hugging Face CLI, where
            the user has logged in through `huggingface-cli login`. If a string is
            specified then it will be used as the token. Defaults to False.

    Returns:
        dict:
            The keys are filterings of the list, which includes all language codes,
            including 'multilingual', as well as 'all'. The values are lists
            of model IDs.
    """
    # Get list of all languages
    all_languages = list(get_all_languages().values())

    # If no languages are specified, then include all languages
    language_list = all_languages if languages is None else languages

    # Form string of languages
    if len(language_list) == 1:
        language_string = language_list[0].name
    else:
        language_list = sorted(language_list, key=lambda x: x.name)
        if {lang.code for lang in language_list} == {
            lang.code for lang in all_languages
        }:
            language_string = "all"
        else:

            # Remove generic 'Norwegian' from the list of languages if both 'Bokmål'
            # and 'Nynorsk' already exist in the list
            if all([lang in language_list for lang in [NO, NB, NN]]):
                language_list = [lang for lang in language_list if lang != NO]

            language_string = (
                f"{', '.join(l.name for l in language_list[:-1])} and "
                f"{language_list[-1].name}"
            )

    # Log fetching message
    logger.info(f"Fetching list of {language_string} models from the Hugging Face Hub.")

    # Initialise the API
    api: HfApi = HfApi()

    # Initialise model lists
    model_lists = defaultdict(list)

    # Do not iterate over all the languages if we are not filtering on language
    language_itr: Sequence[Optional[Language]]
    if {lang.code for lang in language_list} == {lang.code for lang in all_languages}:
        language_itr = [None]
    else:
        language_itr = deepcopy(language_list)

    for language in language_itr:

        # Extract the language code
        language_str: Optional[str]
        if language is not None:
            language_str = language.code
        else:
            language_str = None

        # Fetch the model list
        models: List[ModelInfo] = api.list_models(
            filter=ModelFilter(language=language_str),
            use_auth_token=use_auth_token,
        )

        # Filter the models to only keep the ones with the specified language
        models = [
            model
            for model in models
            if (language is None or language.code in model.tags)
        ]

        # Only keep the models which are not finetuned
        models = [
            model
            for model in models
            if model.pipeline_tag is None
            or model.pipeline_tag
            in {
                "fill-mask",
                "sentence-similarity",
                "feature-extraction",
                "text-generation",
            }
        ]

        # Extract the model IDs
        model_ids: List[str] = [model.modelId for model in models if model.modelId]

        # Remove models that are too large, and thus needs to be specified manually
        large_regex = re.compile(r"(-|_)x+l(arge)?")
        model_ids = [
            model_id
            for model_id in model_ids
            if re.search(large_regex, model_id) is None
        ]

        # Store the model IDs
        model_lists["all"].extend(model_ids)
        if language is not None:
            model_lists[language.code].extend(model_ids)

    # Add multilingual models manually
    multi_models = [
        "bert-base-multilingual-cased",
        "bert-base-multilingual-uncased",
        "cardiffnlp/twitter-xlm-roberta-base",
        "distilbert-base-multilingual-cased",
        "microsoft/infoxlm-base",
        "microsoft/infoxlm-large",
        "microsoft/mdeberta-v3-base",
        "microsoft/xlm-align-base",
        "sentence-transformers/distilbert-multilingual-nli-stsb-quora-ranking",
        "sentence-transformers/distiluse-base-multilingual-cased",
        "sentence-transformers/distiluse-base-multilingual-cased-v1",
        "sentence-transformers/distiluse-base-multilingual-cased-v2",
        "sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2",
        "sentence-transformers/paraphrase-multilingual-mpnet-base-v2",
        "sentence-transformers/paraphrase-xlm-r-multilingual-v1",
        "sentence-transformers/quora-distilbert-multilingual",
        "sentence-transformers/stsb-xlm-r-multilingual",
        "sentence-transformers/use-cmlm-multilingual",
        "studio-ousia/mluke-base",
        "studio-ousia/mluke-large",
        "xlm-roberta-base",
        "xlm-roberta-large",
        "dbmdz/bert-tiny-historic-multilingual-cased",
        "dbmdz/bert-mini-historic-multilingual-cased",
        "dbmdz/bert-base-historic-multilingual-cased",
        "dbmdz/bert-medium-historic-multilingual-cased",
    ]
    model_lists["multilingual"] = multi_models
    model_lists["all"].extend(multi_models)

    # Add fresh models
    fresh_models = ["fresh-xlmr-base", "fresh-electra-small"]
    model_lists["fresh"].extend(fresh_models)
    model_lists["all"].extend(fresh_models)

    # Add some multilingual Danish models manually that have not marked 'da' as their
    # language
    if DA in language_itr:
        multi_da_models: List[str] = [
            "Geotrend/bert-base-en-da-cased",
            "Geotrend/bert-base-25lang-cased",
            "Geotrend/bert-base-en-fr-de-no-da-cased",
            "Geotrend/distilbert-base-en-da-cased",
            "Geotrend/distilbert-base-25lang-cased",
            "Geotrend/distilbert-base-en-fr-de-no-da-cased",
        ]
        model_lists["da"].extend(multi_da_models)
        model_lists["all"].extend(multi_da_models)

    # Add some multilingual Swedish models manually that have not marked 'sv' as their
    # language
    if SV in language_itr:
        multi_sv_models: List[str] = []
        model_lists["sv"].extend(multi_sv_models)
        model_lists["all"].extend(multi_sv_models)

    # Add some multilingual Norwegian models manually that have not marked 'no', 'nb'
    # or 'nn' as their language
    if any(lang in language_itr for lang in [NO, NB, NN]):
        multi_no_models: List[str] = [
            "Geotrend/bert-base-en-no-cased",
            "Geotrend/bert-base-25lang-cased",
            "Geotrend/bert-base-en-fr-de-no-da-cased",
            "Geotrend/distilbert-base-en-no-cased",
            "Geotrend/distilbert-base-25lang-cased",
            "Geotrend/distilbert-base-en-fr-de-no-da-cased",
        ]
        model_lists["no"].extend(multi_no_models)
        model_lists["all"].extend(multi_no_models)

    # Remove duplicates from the lists
    for lang, model_list in model_lists.items():
        model_lists[lang] = list(set(model_list))

    # Remove banned models
    BANNED_MODELS = [
        r"TransQuest/siamesetransquest-da.*",
        r"M-CLIP/.*",
    ]
    for lang, model_list in model_lists.items():
        model_lists[lang] = [
            model
            for model in model_list
            if not any(re.search(regex, model) is not None for regex in BANNED_MODELS)
        ]

    return dict(model_lists)



"""Callbacks for the Hugging Face Trainer."""

from collections.abc import Sized

from tqdm.auto import tqdm
from transformers.trainer_callback import ProgressCallback


class NeverLeaveProgressCallback(ProgressCallback):
    """Progress callback which never leaves the progress bar"""

    def __init__(self, *args, testing: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.testing = testing

    def on_train_begin(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            desc = "Finetuning model"
            self.training_bar = tqdm(
                total=None,
                leave=False,
                desc=desc,
                disable=self.testing,
            )
        self.current_step = 0

    def on_prediction_step(self, args, state, control, eval_dataloader=None, **kwargs):
        if eval_dataloader is None:
            return
        correct_dtype = isinstance(eval_dataloader.dataset, Sized)
        if state.is_local_process_zero and correct_dtype:
            if self.prediction_bar is None:
                desc = "Evaluating model"
                self.prediction_bar = tqdm(
                    total=len(eval_dataloader),
                    leave=False,
                    desc=desc,
                    disable=self.testing,
                )
            self.prediction_bar.update(1)



"""Factory which produces datasets from a configuration."""

from typing import Optional, Type, Union

from .benchmark_dataset import BenchmarkDataset
from .config import BenchmarkConfig, DatasetConfig
from .dataset_configs import get_dataset_config
from .utils import get_class_by_name


class DatasetFactory:
    """Factory which produces datasets from a configuration.

    Args:
        benchmark_config (BenchmarkConfig):
            The benchmark configuration to be used in all datasets constructed.

    Attributes:
        benchmark_config (BenchmarkConfig):
            The benchmark configuration to be used in all datasets constructed.
    """

    def __init__(self, benchmark_config: BenchmarkConfig) -> None:
        self.benchmark_config = benchmark_config

    def build_dataset(self, dataset: Union[str, DatasetConfig]) -> BenchmarkDataset:
        """Build a dataset from a configuration or a name.

        Args:
            dataset (str or DatasetConfig):
                The name of the dataset, or the dataset configuration.

        Returns:
            BenchmarkDataset:
                The benchmark dataset.
        """
        # Get the dataset configuration
        if isinstance(dataset, str):
            dataset_config = get_dataset_config(dataset)
        else:
            dataset_config = dataset

        # Get the benchmark class based on the task
        benchmark_cls: Optional[Type[BenchmarkDataset]] = get_class_by_name(
            [
                dataset_config.name,
                dataset_config.task.name,
                dataset_config.task.supertask,
            ]
        )
        if not benchmark_cls:
            raise ValueError(f"Unknown dataset task: {dataset_config.task}")

        # Create the dataset
        dataset_obj = benchmark_cls(
            dataset_config=dataset_config, benchmark_config=self.benchmark_config
        )

        return dataset_obj



"""Abstract benchmarking dataset class."""

import logging
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from functools import partial
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import evaluate
import numpy as np
import torch
from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict
from datasets.load import load_dataset
from tqdm.auto import tqdm
from transformers.data.data_collator import DataCollator
from transformers.modeling_utils import PreTrainedModel
from transformers.tokenization_utils import PreTrainedTokenizer
from transformers.trainer import Trainer
from transformers.trainer_callback import (
    EarlyStoppingCallback,
    PrinterCallback,
    ProgressCallback,
    TrainerCallback,
)
from transformers.trainer_utils import IntervalStrategy
from transformers.training_args import OptimizerNames, TrainingArguments

from .callbacks import NeverLeaveProgressCallback
from .config import BenchmarkConfig, DatasetConfig, ModelConfig
from .exceptions import InvalidBenchmark
from .hf_hub import get_model_config
from .model_loading import load_model
from .scores import log_scores
from .speed_benchmark import benchmark_speed
from .types import SCORE_DICT
from .utils import (
    block_terminal_output,
    clear_memory,
    enforce_reproducibility,
    handle_error,
)

# Set up logger
logger = logging.getLogger(__name__)


class BenchmarkDataset(ABC):
    """Abstract benchmarking dataset class.

    Args:
        dataset_config (DatasetConfig):
            The configuration of the dataset.
        benchmark_config (BenchmarkConfig):
            The configuration of the benchmark.

    Attributes:
        dataset_config (DatasetConfig):
            The configuration of the dataset.
        benchmark_config (BenchmarkConfig):
            The configuration of the benchmark.
    """

    def __init__(
        self, dataset_config: DatasetConfig, benchmark_config: BenchmarkConfig
    ) -> None:
        """Initialise the dataset.

        Args:
            dataset_config (DatasetConfig):
                The configuration for the dataset.
            benchmark_config (BenchmarkConfig):
                The configuration for the benchmark.
        """
        self.dataset_config = dataset_config
        self.benchmark_config = benchmark_config
        self._metrics = {
            metric_cfg.name: evaluate.load(
                metric_cfg.huggingface_id, cache_dir=self.benchmark_config.cache_dir
            )
            if metric_cfg.huggingface_id != ""
            else None
            for metric_cfg in dataset_config.task.metrics
        }

    def benchmark(  # noqa
        self,
        model_id: str,
    ) -> Tuple[SCORE_DICT, Dict[str, int]]:
        """Benchmark a model.

        Args:
            model_id (str):
                The full Hugging Face Hub path to the pretrained transformer model. The
                specific model version to use can be added after the suffix '@':
                "model_id@v1.0.0". It can be a branch name, a tag name, or a commit id,
                and defaults to the latest version if not specified.

        Returns:
            pair of dicts:
                A pair (score_dict, metadata_dict), with `score_dict` being a
                dictionary containing the scores, and `metadata_dict` being a
                dictionary containing various model metadata, such as the number of
                model parameters, the model's maximum sequence length and the size of
                the model's vocabulary. The keys in `score_dict` are 'raw' and 'total',
                with all the raw scores in the first dictionary and the aggregated
                scores in the second.

        Raises:
            RuntimeError:
                If the extracted framework is not recognized.
        """
        # Fetch the model config
        model_config = get_model_config(
            model_id=model_id, benchmark_config=self.benchmark_config
        )

        # Set random seeds to enforce reproducibility of the randomly initialised
        # weights
        rng = enforce_reproducibility(framework=model_config.framework)

        # Load the model
        tokenizer, model = load_model(
            model_id=model_config.model_id,
            revision=model_config.revision,
            supertask=self.dataset_config.task.supertask,
            num_labels=self.dataset_config.num_labels,
            id2label=self.dataset_config.id2label,
            label2id=self.dataset_config.label2id,
            from_flax=model_config.framework == "jax",
            use_auth_token=self.benchmark_config.use_auth_token,
            cache_dir=self.benchmark_config.cache_dir,
        )

        # Get the metadata
        metadata_dict = self._get_metadata(model=model, tokenizer=tokenizer)

        # Set variable with number of iterations
        num_iter = 10 if not self.benchmark_config.testing else 5

        # Set up progress bar
        itr = tqdm(
            iterable=range(num_iter),
            desc="Benchmarking",
            disable=not self.benchmark_config.progress_bar,
        )

        # If we are running the speed estimation benchmark then call that directly
        if self.dataset_config.task.name == "speed":
            all_scores = benchmark_speed(
                itr=itr,
                tokenizer=tokenizer,
                model=model,
                model_config=model_config,
                dataset_config=self.dataset_config,
                benchmark_config=self.benchmark_config,
            )
            return all_scores, metadata_dict

        # Load the data collator
        data_collator = self._load_data_collator(tokenizer)

        # Load the data
        train, val, test = self._load_data()

        # Get bootstrap sample indices
        test_bidxs = rng.integers(0, len(test), size=(num_iter, len(test)))

        # Get bootstrapped datasets
        tests = [test.select(test_bidxs[idx]) for idx in range(test_bidxs.shape[0])]

        # Set up the preprocessing parameters
        preprocess_params = dict(
            framework="pytorch", config=model.config, tokenizer=tokenizer
        )

        # Prepare the train and validation datasets
        try:
            prepared_train = self._preprocess_data(
                train, split="train", **preprocess_params
            )
            prepared_val = self._preprocess_data(val, split="val", **preprocess_params)
        except ValueError:
            raise InvalidBenchmark(
                "Preprocessing of the training and validation datasets could not be "
                "done."
            )

        # Initialise the `scores` dictionary
        scores: Dict[str, List[Dict[str, float]]] = defaultdict(list)

        bs: int = self.benchmark_config.batch_size
        ga: int = 32 // bs
        for idx in itr:

            # Set variable that tracks whether we need to initialize new models in the
            # `_benchmark_single_iteration` call
            model_already_initialized = idx == 0

            # Clear memory after first iteration
            if not model_already_initialized:
                try:
                    del model
                except UnboundLocalError:
                    pass
                try:
                    del tokenizer
                except UnboundLocalError:
                    pass
                clear_memory()

            while True:

                # Get the boostrapped dataset for this iteration
                test = tests[idx]

                # Prepare the test dataset
                try:
                    prepared_test = self._preprocess_data(
                        test, split="test", **preprocess_params
                    )
                except ValueError:
                    raise InvalidBenchmark(
                        "Preprocessing of the test dataset could not be done."
                    )

                # Re-block terminal output, as it gets unblocked by the `transformers`
                # package before training
                block_terminal_output()

                # Get the training arguments
                training_args = self._get_training_args(iteration_idx=idx)

                # Set the correct batch size and gradient accumulation
                training_args.per_device_train_batch_size = bs
                training_args.per_device_eval_batch_size = bs
                training_args.gradient_accumulation_steps = ga

                itr_scores = self._benchmark_single_iteration(
                    iteration_idx=idx,
                    model_config=model_config,
                    train=train,
                    prepared_train=prepared_train,
                    prepared_val=prepared_val,
                    test=test,
                    prepared_test=prepared_test,
                    data_collator=data_collator,
                    training_args=training_args,
                    tokenizer=tokenizer if model_already_initialized else None,
                    model=model if model_already_initialized else None,
                )

                # If the iteration was successful then break the loop
                if isinstance(itr_scores, dict):
                    break

                # Otherwise we encountered an error, so we have to deal with it and try
                # again
                else:
                    bs = training_args.per_device_train_batch_size
                    ga = training_args.gradient_accumulation_steps

                    bs, ga = handle_error(
                        e=itr_scores,
                        per_device_train_batch_size=bs,
                        gradient_accumulation_steps=ga,
                    )

                    # Clear memory, to avoid memory issues
                    try:
                        del model
                    except UnboundLocalError:
                        pass
                    try:
                        del tokenizer
                    except UnboundLocalError:
                        pass
                    clear_memory()
                    model_already_initialized = False

            if "train" in itr_scores:
                scores["train"].append(itr_scores["train"])
            scores["test"].append(itr_scores["test"])

        all_scores = log_scores(
            dataset_name=self.dataset_config.pretty_name,
            metric_configs=self.dataset_config.task.metrics,
            scores=scores,
            model_id=model_config.model_id,
        )

        return all_scores, metadata_dict

    def _get_metadata(
        self, model: PreTrainedModel, tokenizer: PreTrainedTokenizer
    ) -> Dict[str, int]:

        # Store the number of parameters in the model, the maximum sequence length and
        # the size of the model's vocabulary
        num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        max_seq_length = tokenizer.model_max_length
        if hasattr(model.config, "vocab_size"):
            vocab_size = model.config.vocab_size
        elif hasattr(tokenizer, "vocab_size"):
            vocab_size = tokenizer.vocab_size
        else:
            vocab_size = -1

        # Store the metadata in a dictionary
        metadata_dict = dict(
            num_model_parameters=num_params,
            max_sequence_length=max_seq_length,
            vocabulary_size=vocab_size,
        )

        # Log the metadata
        logger.info(
            f"The model has {num_params:,} parameters, a vocabulary size of "
            f"{vocab_size:,} and a maximum sequence length of {max_seq_length:,}."
        )

        return metadata_dict

    def _get_training_args(self, iteration_idx: int) -> TrainingArguments:

        # Set the logging strategy
        if self.benchmark_config.verbose:
            logging_strategy = IntervalStrategy.STEPS
        else:
            logging_strategy = IntervalStrategy.NO

        # Set seed variable
        seed = 4242 + iteration_idx

        # Initialise training arguments
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            training_args = TrainingArguments(
                output_dir=self.benchmark_config.cache_dir,
                evaluation_strategy=IntervalStrategy.STEPS,
                logging_strategy=logging_strategy,
                save_strategy=IntervalStrategy.STEPS,
                eval_steps=30,
                logging_steps=30,
                save_steps=30,
                max_steps=10_000 if not self.benchmark_config.testing else 10,
                report_to=[],
                save_total_limit=1,
                per_device_train_batch_size=self.benchmark_config.batch_size,
                per_device_eval_batch_size=self.benchmark_config.batch_size,
                learning_rate=2e-5,
                warmup_ratio=0.01,
                gradient_accumulation_steps=1,
                load_best_model_at_end=True,
                optim=OptimizerNames.ADAMW_TORCH,
                seed=seed,
                use_mps_device=torch.backends.mps.is_available(),
                fp16=False,
            )

        # Manually set `disable_tqdm` to `False` if `progress_bar` is `True`
        if self.benchmark_config.progress_bar:
            training_args.disable_tqdm = False

        return training_args

    def _load_data(self):

        # Download dataset from the HF Hub
        dataset_dict = load_dataset(
            path=self.dataset_config.huggingface_id,
            use_auth_token=self.benchmark_config.use_auth_token,
            cache_dir=self.benchmark_config.cache_dir,
        )

        # If the dataset turns out not to be a DatasetDict, then we raise an error
        if not isinstance(dataset_dict, DatasetDict):
            raise InvalidBenchmark(
                f"Expected `dataset_dict` to be a `DatasetDict`, but got "
                f"{type(dataset_dict)}."
            )

        # Remove all other keys than 'train', 'val' and 'test'
        dataset_dict = DatasetDict(
            {key: dataset_dict[key] for key in ["train", "val", "test"]}
        )

        # Process the datasets
        dataset_dict = self._process_data(dataset_dict)

        # Extract the dataset splits
        train = dataset_dict["train"]
        val = dataset_dict["val"]
        test = dataset_dict["test"]

        # Remove empty examples from the datasets
        if "tokens" in train.features:
            train = train.filter(lambda x: len(x["tokens"]) > 0)
            val = val.filter(lambda x: len(x["tokens"]) > 0)
            test = test.filter(lambda x: len(x["tokens"]) > 0)
        elif "doc" in train.features:
            train = train.filter(lambda x: len(x["doc"]) > 0)
            val = val.filter(lambda x: len(x["doc"]) > 0)
            test = test.filter(lambda x: len(x["doc"]) > 0)

        # If we are testing then truncate the test set
        if self.benchmark_config.testing:
            test = test.select(range(128))

        return train, val, test

    def _benchmark_single_iteration(
        self,
        iteration_idx: int,
        model_config: ModelConfig,
        train: Dataset,
        prepared_train: Dataset,
        prepared_val: Dataset,
        test: Dataset,
        prepared_test: Dataset,
        data_collator: DataCollator,
        training_args: TrainingArguments,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        model: Optional[PreTrainedModel] = None,
    ) -> Union[Dict[str, Dict[str, float]], Exception]:
        """Run a single iteration of a benchmark.

        Args:
            iteration_idx (int):
                The index of the iteration.
            model_config (ModelConfig):
                The model configuration.
            train (Dataset):
                The original training dataset.
            prepared_train (Dataset):
                The prepared training dataset.
            prepared_val (Dataset):
                The prepared validation dataset.
            test (Dataset):
                The original test dataset.
            prepared_test (Dataset):
                The prepared test dataset.
            data_collator (DataCollator):
                The data collator.
            training_args (TrainingArguments):
                The training arguments.
            tokenizer (PreTrainedTokenizer or None, optional):
                The tokenizer to use in the benchmark. If None then a new tokenizer
                will be loaded. Defaults to None.
            model (PreTrainedModel or None, optional):
                The model to use in the benchmark. If None then a new model will be
                loaded. Defaults to None.

        Returns:
            dict or Exception:
                A dictionary containing the scores for the current iteration, with keys
                `train` and `test`. If an exception is raised, then the exception is
                returned.
        """
        scores: Dict[str, Dict[str, float]] = dict()
        try:
            # Set random seeds to enforce reproducibility of the randomly initialised
            # weights
            seed = 4242 + iteration_idx
            enforce_reproducibility(framework=model_config.framework, seed=seed)

            # Reinitialise a new model
            if tokenizer is None or model is None:
                tokenizer, model = load_model(
                    model_id=model_config.model_id,
                    revision=model_config.revision,
                    supertask=self.dataset_config.task.supertask,
                    num_labels=self.dataset_config.num_labels,
                    label2id=self.dataset_config.label2id,
                    id2label=self.dataset_config.id2label,
                    from_flax=model_config.framework == "jax",
                    use_auth_token=self.benchmark_config.use_auth_token,
                    cache_dir=self.benchmark_config.cache_dir,
                )

            # Initialise compute_metrics function
            compute_metrics = partial(
                self._compute_metrics, id2label=model.config.id2label
            )

            # Initialise early stopping callback
            early_stopping = EarlyStoppingCallback(early_stopping_patience=2)

            # Initialise trainer
            trainer = self._get_trainer(
                model=model,
                args=training_args,
                train_dataset=prepared_train,
                eval_dataset=prepared_val,
                tokenizer=tokenizer,
                data_collator=data_collator,
                compute_metrics=compute_metrics,
                callbacks=[early_stopping],
            )

            # Remove trainer logging if not in verbose mode
            if not self.benchmark_config.verbose:
                trainer.log = lambda logs: None

            # Re-block terminal output, as it gets unblocked by the `transformers`
            # package before training
            block_terminal_output()

            # Remove the callback which prints the scores after each evaluation
            if not self.benchmark_config.verbose:
                trainer.remove_callback(PrinterCallback)

            # Remove the progress bar callback
            trainer.remove_callback(ProgressCallback)

            # Add the custom progress callback if `progress_bar` is True
            if self.benchmark_config.progress_bar:
                trainer.add_callback(NeverLeaveProgressCallback)

            # Finetune the model
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                trainer.train()

            # Log training scores and save the state
            if self.benchmark_config.evaluate_train:
                train_scores = self._evaluate_dataset(
                    dataset=train,
                    prepared_dataset=prepared_train,
                    metric_key_prefix="train",
                    trainer=trainer,
                )
                scores["train"] = train_scores

            # Log test scores
            test_scores = self._evaluate_dataset(
                dataset=test,
                prepared_dataset=prepared_test,
                metric_key_prefix="test",
                trainer=trainer,
            )
            scores["test"] = test_scores

            # Return the scores
            return scores

        except (RuntimeError, ValueError, IndexError) as e:
            try:
                del model
            except UnboundLocalError:
                pass
            try:
                del tokenizer
            except UnboundLocalError:
                pass
            clear_memory()
            return e

    def __call__(self, *args, **kwargs):
        return self.benchmark(*args, **kwargs)

    def _get_trainer(
        self,
        model: PreTrainedModel,
        args: TrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        tokenizer: PreTrainedTokenizer,
        data_collator: DataCollator,
        compute_metrics: Callable,
        callbacks: List[TrainerCallback],
    ) -> Trainer:
        """Get a Trainer object.

        Args:
            model (PreTrainedModel):
                The model to finetune.
            args (TrainingArguments):
                The training arguments.
            train_dataset (Dataset):
                The training dataset.
            eval_dataset (Dataset):
                The evaluation dataset.
            tokenizer (PreTrainedTokenizer):
                The tokenizer.
            data_collator (DataCollator):
                The data collator.
            compute_metrics (Callable):
                The function used to compute the metrics.
            callbacks (list of TrainerCallback):
                The callbacks to use.

        Returns:
            Trainer:
                The Trainer object.
        """
        return Trainer(
            model=model,
            args=args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            data_collator=data_collator,
            compute_metrics=compute_metrics,
            callbacks=callbacks,
        )

    def _evaluate_dataset(
        self,
        trainer: Trainer,
        dataset: Dataset,
        prepared_dataset: Dataset,
        metric_key_prefix: str,
    ) -> Dict[str, float]:
        """Evaluate a dataset.

        Args:
            trainer (Trainer):
                The trainer.
            dataset (Dataset):
                The original dataset.
            prepared_dataset (Dataset):
                The prepared dataset.
            metric_key_prefix (str):
                The prefix to use for the metric keys.

        Returns:
            dict:
                The scores.
        """
        return trainer.evaluate(
            eval_dataset=prepared_dataset, metric_key_prefix=metric_key_prefix
        )

    def _process_data(self, dataset_dict: DatasetDict) -> DatasetDict:
        """Process the data.

        Args:
            dataset_dict (DatasetDict):
                The dataset dictionary.

        Returns:
            DatasetDict:
                The processed dataset dictionary.
        """
        return dataset_dict

    @abstractmethod
    def _preprocess_data(self, dataset: Dataset, **kwargs) -> Dataset:
        """Preprocess a dataset.

        Args:
            dataset (Hugging Face dataset):
                The dataset to preprocess.
            kwargs:
                Extra keyword arguments containing objects used in preprocessing the
                dataset.

        Returns:
            Hugging Face dataset:
                The preprocessed dataset.
        """
        pass

    @abstractmethod
    def _load_data_collator(
        self, tokenizer: Optional[PreTrainedTokenizer] = None
    ) -> DataCollator:
        """Load the data collator used to prepare samples during finetuning.

        Args:
            tokenizer (PreTrainedTokenizer or None, optional):
                A pretrained tokenizer. Can be None if the tokenizer is not used in the
                initialisation of the data collator. Defaults to None.

        Returns:
            Hugging Face data collator:
                The data collator.
        """
        pass

    def _compute_metrics(
        self,
        model_outputs_and_labels: Tuple[Sequence, Sequence],
        id2label: Optional[Sequence[str]] = None,
    ) -> Dict[str, float]:
        """Compute the metrics needed for evaluation.

        Args:
            model_outputs_and_labels (pair of sequences):
                The first sequence contains the model outputs and the second sequence
                contains the true labels.
            id2label (list or None, optional):
                Conversion of indices to labels. Defaults to None.

        Returns:
            dict:
                A dictionary with the names of the metrics as keys and the metric
                values as values.
        """
        model_outputs, labels = model_outputs_and_labels

        model_output_dtype = np.asarray(model_outputs).dtype
        if model_output_dtype in [np.float16, np.float32, np.float64]:
            predictions = np.asarray(model_outputs).argmax(axis=-1)
        else:
            predictions = model_outputs

        results: Dict[str, float] = dict()
        for cfg in self.dataset_config.task.metrics:
            metric = self._metrics[cfg.name]
            score_dict: Union[None, Dict[str, float]] = metric.compute(
                predictions=predictions,
                references=labels,
                **cfg.compute_kwargs,
            )
            if score_dict is not None:
                scores = score_dict[cfg.results_key]
                results[cfg.name] = scores
        return results



"""All dataset configurations used in ScandEval."""

from typing import Dict

from .config import DatasetConfig
from .dataset_tasks import LA, NER, QA, SENT, SPEED
from .languages import DA, FO, IS, NB, NN, SV, get_all_languages


def get_all_dataset_configs() -> Dict[str, DatasetConfig]:
    """Get a mapping of all the dataset configurations.

    Returns:
        dict:
            A mapping between names of datasets and their configurations.
    """
    return {
        cfg.name: cfg for cfg in globals().values() if isinstance(cfg, DatasetConfig)
    }


def get_dataset_config(dataset_name: str) -> DatasetConfig:
    """Get the dataset configuration for a dataset.

    Args:
        dataset_name (str):
            The name of the dataset.

    Returns:
        DatasetConfig:
            The dataset configuration.

    Raises:
        ValueError:
            If the dataset is not found.
    """
    # Get mapping of all dataset configs
    dataset_configs = get_all_dataset_configs()

    # If there are no matches for the dataset name, raise an error
    if dataset_name not in dataset_configs:
        raise ValueError(f"No dataset config found for dataset {dataset_name}.")

    # Otherwise, return the dataset configuration
    return dataset_configs[dataset_name]


SWEREC_CONFIG = DatasetConfig(
    name="swerec",
    pretty_name="the truncated version of SweReC",
    huggingface_id="ScandEval/swerec-mini",
    task=SENT,
    languages=[SV],
)


ANGRY_TWEETS_CONFIG = DatasetConfig(
    name="angry-tweets",
    pretty_name="the truncated version of AngryTweets",
    huggingface_id="ScandEval/angry-tweets-mini",
    task=SENT,
    languages=[DA],
)


NOREC_CONFIG = DatasetConfig(
    name="norec",
    pretty_name="the truncated version of NoReC",
    huggingface_id="ScandEval/norec-mini",
    task=SENT,
    languages=[NB, NN],
)


SUC3_CONFIG = DatasetConfig(
    name="suc3",
    pretty_name="the truncated version of SUC 3.0",
    huggingface_id="ScandEval/suc3-mini",
    task=NER,
    languages=[SV],
)


DANE_CONFIG = DatasetConfig(
    name="dane",
    pretty_name="the truncated version of DaNE",
    huggingface_id="ScandEval/dane-mini",
    task=NER,
    languages=[DA],
)


NORNE_NB_CONFIG = DatasetConfig(
    name="norne-nb",
    pretty_name="the truncated version of the Bokmål part of NorNE",
    huggingface_id="ScandEval/norne-nb-mini",
    task=NER,
    languages=[NB],
)


NORNE_NN_CONFIG = DatasetConfig(
    name="norne-nn",
    pretty_name="the truncated version of the Nynorsk part of NorNE",
    huggingface_id="ScandEval/norne-nn-mini",
    task=NER,
    languages=[NN],
)


MIM_GOLD_NER_CONFIG = DatasetConfig(
    name="mim-gold-ner",
    pretty_name="the truncated version of MIM-GOLD-NER",
    huggingface_id="ScandEval/mim-gold-ner-mini",
    task=NER,
    languages=[IS],
)


WIKIANN_FO_CONFIG = DatasetConfig(
    name="wikiann-fo",
    pretty_name="the truncated version of the Faroese part of WikiANN",
    huggingface_id="ScandEval/wikiann-fo-mini",
    task=NER,
    languages=[FO],
)


SCALA_SV_CONFIG = DatasetConfig(
    name="scala-sv",
    pretty_name="The Swedish part of ScaLA",
    huggingface_id="ScandEval/scala-sv",
    task=LA,
    languages=[SV],
)


SCALA_DA_CONFIG = DatasetConfig(
    name="scala-da",
    pretty_name="the Danish part of ScaLA",
    huggingface_id="ScandEval/scala-da",
    task=LA,
    languages=[DA],
)


SCALA_NB_CONFIG = DatasetConfig(
    name="scala-nb",
    pretty_name="the Bokmål part of ScaLA",
    huggingface_id="ScandEval/scala-nb",
    task=LA,
    languages=[NB],
)


SCALA_NN_CONFIG = DatasetConfig(
    name="scala-nn",
    pretty_name="the Nynorsk part of ScaLA",
    huggingface_id="ScandEval/scala-nn",
    task=LA,
    languages=[NN],
)


SCALA_IS_CONFIG = DatasetConfig(
    name="scala-is",
    pretty_name="the Icelandic part of ScaLA",
    huggingface_id="ScandEval/scala-is",
    task=LA,
    languages=[IS],
)


SCALA_FO_CONFIG = DatasetConfig(
    name="scala-fo",
    pretty_name="the Faroese part of ScaLA",
    huggingface_id="ScandEval/scala-fo",
    task=LA,
    languages=[FO],
)


SCANDIQA_DA_CONFIG = DatasetConfig(
    name="scandiqa-da",
    pretty_name="the Danish part of the truncated version of ScandiQA",
    huggingface_id="ScandEval/scandiqa-da-mini",
    task=QA,
    languages=[DA],
)


SCANDIQA_NO_CONFIG = DatasetConfig(
    name="scandiqa-no",
    pretty_name="the Norwegian part of truncated version of ScandiQA",
    huggingface_id="ScandEval/scandiqa-no-mini",
    task=QA,
    languages=[NB, NN],
)


SCANDIQA_SV_CONFIG = DatasetConfig(
    name="scandiqa-sv",
    pretty_name="the Swedish part of the truncated version of ScandiQA",
    huggingface_id="ScandEval/scandiqa-sv-mini",
    task=QA,
    languages=[SV],
)


SPEED_CONFIG = DatasetConfig(
    name="speed",
    pretty_name="the speed estimation benchmark",
    huggingface_id="",
    task=SPEED,
    languages=list(get_all_languages().values()),
)



"""Benchmarking model inference speed."""

from collections import defaultdict
from typing import Dict, List, Union

import pyinfer
import torch
from tqdm.auto import tqdm
from transformers.modeling_utils import PreTrainedModel
from transformers.tokenization_utils import PreTrainedTokenizer

from .benchmark_dataset import log_scores
from .config import BenchmarkConfig, DatasetConfig, ModelConfig
from .exceptions import InvalidBenchmark
from .model_loading import load_model
from .types import SCORE_DICT


def benchmark_speed(
    itr: tqdm,
    tokenizer: PreTrainedTokenizer,
    model: PreTrainedModel,
    model_config: ModelConfig,
    dataset_config: DatasetConfig,
    benchmark_config: BenchmarkConfig,
) -> SCORE_DICT:
    """Benchmark model inference speed.

    Args:
        itr (tqdm):
            tqdm iterator.
        tokenizer (PreTrainedTokenizer):
            Tokenizer to use.
        model (PreTrainedModel):
            Model to use.
        model_config (ModelConfig):
            Model configuration.
        dataset_config (DatasetConfig):
            Dataset configuration.
        benchmark_config (BenchmarkConfig):
            Benchmark configuration.

    Returns:
        SCORE_DICT:
            Dictionary of scores.
    """
    # Initialise the `scores` dictionary
    scores: Dict[str, List[Dict[str, float]]] = defaultdict(list)

    for _ in itr:

        # Run the speed benchmark
        itr_scores = benchmark_speed_single_iteration(
            tokenizer=tokenizer,
            model=model,
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
        )

        # If the iteration was unsuccessful then raise an error
        if isinstance(itr_scores, Exception):
            raise InvalidBenchmark(f"Speed benchmark failed with error: {itr_scores!r}")

        # Otherwise, append the scores to the list and log the result
        else:
            scores["test"].append(itr_scores["test"])
            if benchmark_config.evaluate_train:
                scores["train"].append(itr_scores["train"])
            if benchmark_config.verbose:
                print(itr_scores)

    all_scores = log_scores(
        dataset_name=dataset_config.pretty_name,
        metric_configs=dataset_config.task.metrics,
        scores=scores,
        model_id=model_config.model_id,
    )

    return all_scores


def benchmark_speed_single_iteration(
    tokenizer: PreTrainedTokenizer,
    model: PreTrainedModel,
    model_config: ModelConfig,
    dataset_config: DatasetConfig,
    benchmark_config: BenchmarkConfig,
) -> Union[Dict[str, Dict[str, float]], Exception]:
    """Run a single iteration of the speed benchmark.

    Args:
        tokenizer (PreTrainedTokenizer):
            The tokenizer to use in the benchmark.
        model (PreTrainedModel):
            The model to use in the benchmark.
        model_config (ModelConfig):
            The model configuration.
        dataset_config (DatasetConfig):
            The dataset configuration.
        benchmark_config (BenchmarkConfig):
            The benchmark configuration.

    Returns:
        dict or Exception:
            A dictionary containing the scores for the current iteration, with keys
            `train` and `test`. If an exception is raised, then the exception is
            returned.
    """
    scores: Dict[str, Dict[str, float]] = dict()
    try:

        # Reinitialise a new model
        if tokenizer is None or model is None:
            tokenizer, model = load_model(
                model_id=model_config.model_id,
                revision=model_config.revision,
                supertask=dataset_config.task.supertask,
                num_labels=dataset_config.num_labels,
                label2id=dataset_config.label2id,
                id2label=dataset_config.id2label,
                from_flax=model_config.framework == "jax",
                use_auth_token=benchmark_config.use_auth_token,
                cache_dir=benchmark_config.cache_dir,
            )

        # Ensure that the model is on the CPU
        model.cpu()

        # Create a dummy document
        doc = "This is a dummy document. " * 100

        def predict(docs: List[str]) -> None:
            """Function used to benchmark inference speed of the model."""

            # Raise an error if the tokenizer or model is undefined
            if tokenizer is None or model is None:
                raise ValueError("Tokenizer and model must not be None.")

            # Tokenize the document
            inputs = tokenizer(
                docs,
                padding=True,
                truncation=True,
                return_tensors="pt",
            )

            # Run inference with the model
            with torch.no_grad():
                model(**inputs)

        # Do a warmup run
        pyinfer.InferenceReport(model=predict, inputs=doc, n_iterations=10).run(
            print_report=False
        )

        # Initialise the speed benchmark
        speed_benchmark = pyinfer.InferenceReport(
            model=predict,
            inputs=doc,
            n_iterations=100,
        )

        # Run the speed benchmark
        speed_scores = speed_benchmark.run(print_report=False)

        # Close the speed benchmark
        del speed_benchmark

        # Store the scores
        scores["test"] = {"test_speed": speed_scores["Infer(p/sec)"]}
        if benchmark_config.evaluate_train:
            scores["train"] = {"train_speed": speed_scores["Infer(p/sec)"]}

        # Return the scores
        return scores

    except (RuntimeError, ValueError, IndexError) as e:
        return e



"""Functions related to the loading of models."""

import warnings
from json import JSONDecodeError
from typing import Dict, List, Tuple, Type, Union

from transformers import PreTrainedModel
from transformers.configuration_utils import PretrainedConfig
from transformers.models.auto.configuration_auto import AutoConfig
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.models.electra.modeling_electra import (
    ElectraForQuestionAnswering,
    ElectraForSequenceClassification,
    ElectraForTokenClassification,
)
from transformers.models.xlm_roberta.modeling_xlm_roberta import (
    XLMRobertaForQuestionAnswering,
    XLMRobertaForSequenceClassification,
    XLMRobertaForTokenClassification,
)
from transformers.tokenization_utils import PreTrainedTokenizer

from .exceptions import InvalidBenchmark
from .utils import block_terminal_output, get_class_by_name


def load_model(
    model_id: str,
    revision: str,
    supertask: str,
    num_labels: int,
    id2label: List[str],
    label2id: Dict[str, int],
    from_flax: bool,
    use_auth_token: Union[bool, str],
    cache_dir: str,
) -> Tuple[PreTrainedTokenizer, PreTrainedModel]:
    """Load a model.

    Args:
        model_id (str):
            The Hugging Face ID of the model.
        revision (str):
            The specific version of the model. Can be a branch name, a tag name, or a
            commit hash.
        supertask (str):
            The supertask of the task to benchmark the model on.
        num_labels (int):
            The number of labels in the dataset.
        id2label (list of str):
            The mapping from ID to label.
        label2id (dict of str to int):
            The mapping from label to ID.
        from_flax (bool):
            Whether the model is a Flax model.
        use_auth_token (bool or str):
            Whether to use an authentication token to access the model. If a boolean
            value is specified then it is assumed that the user is logged in to the
            Hugging Face CLI, and if a string is specified then it is used as the
            token.
        cache_dir (str):
            The directory to cache the model in.

    Returns:
        pair of (tokenizer, model):
            The tokenizer and model.

    Raises:
        RuntimeError:
            If the framework is not recognized.
    """
    config: PretrainedConfig
    block_terminal_output()

    try:
        # If the model ID specifies a fresh model, then load that.
        if model_id.startswith("fresh"):
            model_cls, model_id = load_fresh_model_class(
                model_id=model_id, supertask=supertask
            )
            config = AutoConfig.from_pretrained(
                model_id,
                use_auth_token=use_auth_token,
                num_labels=num_labels,
                id2label=id2label,
                label2id=label2id,
                cache_dir=cache_dir,
            )
            model = model_cls(config)

        # Otherwise load the pretrained model
        else:
            try:
                config = AutoConfig.from_pretrained(
                    model_id,
                    revision=revision,
                    use_auth_token=use_auth_token,
                    num_labels=num_labels,
                    id2label=id2label,
                    label2id=label2id,
                    cache_dir=cache_dir,
                )
            except KeyError as e:
                key = e.args[0]
                raise InvalidBenchmark(
                    f"The model config for the mmodel {model_id!r} could not be "
                    f"loaded, as the key {key!r} was not found in the config."
                )

            # Get the model class associated with the supertask
            model_cls_or_none: Union[None, Type[PreTrainedModel]] = get_class_by_name(
                class_name=f"auto-model-for-{supertask}",
                module_name="transformers",
            )

            # If the model class could not be found then raise an error
            if not model_cls_or_none:
                raise InvalidBenchmark(
                    f"The supertask '{supertask}' does not correspond to a Hugging Face"
                    " AutoModel type (such as `AutoModelForSequenceClassification`)."
                )

            # If the model is a DeBERTaV2 model then we ensure that
            # `pooler_hidden_size` is the same size as `hidden_size`
            if config.model_type == "deberta-v2":
                config.pooler_hidden_size = config.hidden_size

            # If we are benchmarking a question answering model then we need to set
            # the type_vocab_size to 2
            if supertask == "question-answering":
                config.type_vocab_size = 2

            # Load the model
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UserWarning)
                model_or_tuple = model_cls_or_none.from_pretrained(
                    model_id,
                    revision=revision,
                    use_auth_token=use_auth_token,
                    config=config,
                    cache_dir=cache_dir,
                    from_flax=from_flax,
                    ignore_mismatched_sizes=True,
                )
            if isinstance(model_or_tuple, tuple):
                model = model_or_tuple[0]
            else:
                model = model_or_tuple

    except (OSError, ValueError) as e:

        # Deal with the case where the checkpoint is incorrect
        if "checkpoint seems to be incorrect" in str(e):
            raise InvalidBenchmark(
                f"The model {model_id!r} has an incorrect checkpoint."
            )

        # Otherwise raise a more generic error
        raise InvalidBenchmark(
            f"The model {model_id} either does not exist on the Hugging Face Hub, or "
            "it has no frameworks registered, or it is a private model. If it *does* "
            "exist on the Hub and is a public model then please ensure that it has a "
            "framework registered. If it is a private model then enable the "
            "`--use-auth-token` flag and make sure that you are logged in to the Hub "
            "via the `huggingface-cli login` command."
        )

    # If the model is a subclass of a RoBERTa model then we have to add a prefix
    # space to the tokens, by the way the model is constructed.
    prefix_models = ["Roberta", "GPT", "Deberta"]
    prefix = any(model_type in type(model).__name__ for model_type in prefix_models)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        try:
            tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(
                model_id,
                revision=revision,
                use_auth_token=use_auth_token,
                add_prefix_space=prefix,
                cache_dir=cache_dir,
                use_fast=True,
                verbose=False,
            )
        except JSONDecodeError:
            raise InvalidBenchmark(f"Could not load tokenizer for model {model_id!r}.")

    # Fix the model and the tokenizer
    model, tokenizer = fix_model_and_tokenizer(model, tokenizer)

    return tokenizer, model


def load_fresh_model_class(
    model_id: str, supertask: str
) -> Tuple[Type[PreTrainedModel], str]:
    """Load a fresh model class.

    Args:
        model_id (str):
            The Hugging Face ID of the model.
        supertask (str):
            The supertask of the task to benchmark the model on.

    Returns:
        pair of Type[PreTrainedModel] and str:
            The model class and the pretrained model ID.
    """
    if model_id == "fresh-xlmr-base":
        model_id = "xlm-roberta-base"
        if supertask == "sequence-classification":
            model_cls = XLMRobertaForSequenceClassification
        elif supertask == "token-classification":
            model_cls = XLMRobertaForTokenClassification
        elif supertask == "question-answering":
            model_cls = XLMRobertaForQuestionAnswering
        else:
            raise InvalidBenchmark(
                f"Supertask {supertask} is not supported for model {model_id}"
            )

    elif model_id == "fresh-electra-small":
        model_id = "google/electra-small-discriminator"
        if supertask == "sequence-classification":
            model_cls = ElectraForSequenceClassification
        elif supertask == "token-classification":
            model_cls = ElectraForTokenClassification
        elif supertask == "question-answering":
            model_cls = ElectraForQuestionAnswering
        else:
            raise InvalidBenchmark(
                f"Supertask {supertask} is not supported for model {model_id}"
            )

    else:
        raise InvalidBenchmark(
            f"A fresh model was chosen, `{model_id}`, but it was not recognized."
        )

    return model_cls, model_id


def fix_model_and_tokenizer(
    model: PreTrainedModel, tokenizer: PreTrainedTokenizer
) -> Tuple[PreTrainedModel, PreTrainedTokenizer]:
    """Fixes the model and tokenizer to be compatible with the benchmarking framework.

    Args:
        model (PreTrainedModel):
            The model to fix.
        tokenizer (PreTrainedTokenizer):
            The tokenizer to fix.

    Returns:
        pair of (model, tokenizer):
            The fixed model and tokenizer.
    """
    # Set the maximal length of the tokenizer to the model's maximal length. This is
    # required for proper truncation
    if (
        not hasattr(tokenizer, "model_max_length")
        or tokenizer.model_max_length > 100_000
    ):

        if hasattr(tokenizer, "max_model_input_sizes"):
            all_max_lengths = tokenizer.max_model_input_sizes.values()
            if len(list(all_max_lengths)) > 0:
                min_max_length = min(list(all_max_lengths))
                tokenizer.model_max_length = min_max_length
            else:
                tokenizer.model_max_length = 512
        else:
            tokenizer.model_max_length = 512

    # If there is a mismatch between the vocab size according to the tokenizer and
    # the vocab size according to the model, we raise an error
    if hasattr(model.config, "vocab_size") and hasattr(tokenizer, "vocab_size"):
        if model.config.vocab_size < tokenizer.vocab_size:
            raise InvalidBenchmark(
                "The vocab size of the tokenizer is larger than the vocab size of the "
                "model. This is not supported."
            )

    # If the tokenizer does not have a padding token (e.g. GPT-2), we use find a
    # suitable padding token and set it
    if tokenizer.pad_token is None:
        if tokenizer.eos_token is not None:
            tokenizer.padding_side = "left"
            tokenizer.pad_token = tokenizer.eos_token
            model.config.pad_token_id = tokenizer.pad_token_id
        elif tokenizer.sep_token is not None:
            tokenizer.padding_side = "left"
            tokenizer.pad_token = tokenizer.sep_token
            model.config.pad_token_id = tokenizer.pad_token_id
        else:
            raise InvalidBenchmark(
                "The tokenizer does not have a padding token and does not have a "
                "SEP token or EOS token to use as a padding token."
            )

    return model, tokenizer



"""Sequence classification benchmark dataset."""

import logging
from functools import partial
from typing import Optional

from datasets.arrow_dataset import Dataset
from transformers import BatchEncoding
from transformers.data.data_collator import DataCollator, DataCollatorWithPadding
from transformers.tokenization_utils import PreTrainedTokenizer

from scandeval.utils import get_special_token_metadata

from .benchmark_dataset import BenchmarkDataset
from .exceptions import InvalidBenchmark

logger = logging.getLogger(__name__)


class SequenceClassification(BenchmarkDataset):
    """Sequence classification benchmark dataset.

    Args:
        dataset_config (DatasetConfig):
            The dataset configuration.
        benchmark_config (BenchmarkConfig):
            The benchmark configuration.

    Attributes:
        dataset_config (DatasetConfig):
            The configuration of the dataset.
        benchmark_config (BenchmarkConfig):
            The configuration of the benchmark.
    """

    def _preprocess_data(self, dataset: Dataset, **kwargs) -> Dataset:
        """Preprocess a dataset by tokenizing and aligning the labels.

        Args:
            dataset (Hugging Face dataset):
                The dataset to preprocess.
            kwargs:
                Extra keyword arguments containing objects used in preprocessing the
                dataset.

        Returns:
            Hugging Face dataset:
                The preprocessed dataset.
        """
        tokenizer: PreTrainedTokenizer = kwargs["tokenizer"]

        # Extract special token metadata from the tokenizer
        special_token_metadata = get_special_token_metadata(tokenizer=tokenizer)
        has_cls_token = special_token_metadata["has_cls_token"]
        has_sep_token = special_token_metadata["has_sep_token"]
        cls_token = special_token_metadata["cls_token"]
        sep_token = special_token_metadata["sep_token"]

        def tokenise(examples: dict) -> BatchEncoding:

            # If the tokenizer is not adding special tokens, then we add them manually
            if (
                not has_cls_token
                and not has_sep_token
                and cls_token is not None
                and sep_token is not None
            ):
                examples["text"] = [
                    f"{cls_token}{doc}{sep_token}" for doc in examples["text"]
                ]

            return tokenizer(examples["text"], truncation=True, padding=False)

        tokenised = dataset.map(tokenise, batched=True, load_from_cache_file=False)

        numericalise = partial(
            self._create_numerical_labels, label2id=kwargs["config"].label2id
        )
        preprocessed = tokenised.map(
            numericalise, batched=True, load_from_cache_file=False
        )

        return preprocessed.remove_columns(["text"])

    def _create_numerical_labels(self, examples: dict, label2id: dict) -> dict:
        try:
            examples["label"] = [label2id[lbl.upper()] for lbl in examples["label"]]
        except KeyError:
            raise InvalidBenchmark(
                f"One of the labels in the dataset, {examples['label'].upper()}, does "
                f"not occur in the label2id dictionary {label2id}."
            )
        return examples

    def _load_data_collator(
        self, tokenizer: Optional[PreTrainedTokenizer] = None
    ) -> DataCollator:
        """Load the data collator used to prepare samples during finetuning.

        Args:
            tokenizer (Hugging Face tokenizer or None, optional):
                A pretrained tokenizer. Can be None if the tokenizer is not used in the
                initialisation of the data collator. Defaults to None.

        Returns:
            Hugging Face data collator:
                The data collator.
        """
        return DataCollatorWithPadding(tokenizer, padding="longest")



"""Command-line interface for benchmarking."""

from typing import List, Tuple, Union

import click

from .benchmarker import Benchmarker
from .dataset_configs import get_all_dataset_configs
from .dataset_tasks import get_all_dataset_tasks
from .languages import get_all_languages


@click.command()
@click.option(
    "--model-id",
    "-m",
    default=None,
    show_default=True,
    multiple=True,
    help="""The Hugging Face model ID of the model(s) to be benchmarked. If not
    specified then all models will be benchmarked, filtered by `model_language` and
    `model_task`. The specific model version to use can be added after the suffix "@":
    "<model_id>@v1.0.0". It can be a branch name, a tag name, or a commit id, and
    defaults to the latest version if not specified.""",
)
@click.option(
    "--dataset",
    "-d",
    default=None,
    show_default=True,
    multiple=True,
    type=click.Choice(list(get_all_dataset_configs().keys())),
    help="""The name of the benchmark dataset. If not specified then all datasets will
    be benchmarked, filtered by `dataset_language` and `dataset_task`.""",
)
@click.option(
    "--language",
    "-l",
    default=["da", "sv", "no"],
    show_default=True,
    multiple=True,
    metavar="ISO 639-1 LANGUAGE CODE",
    type=click.Choice(["all"] + list(get_all_languages().keys())),
    help="""The languages to benchmark, both for models and datasets. Only relevant if
    `model-id` and `dataset` have not both been specified. If "all" then all models
    will be benchmarked on all datasets.""",
)
@click.option(
    "--model-language",
    "-ml",
    default=None,
    show_default=True,
    multiple=True,
    metavar="ISO 639-1 LANGUAGE CODE",
    type=click.Choice(["all"] + list(get_all_languages().keys())),
    help="""The model languages to benchmark. Only relevant if `model-id` has not been
    specified. If "all" then all models will be benchmarked. If not specified then this
    will use the `language` value.""",
)
@click.option(
    "--dataset-language",
    "-dl",
    default=None,
    show_default=True,
    multiple=True,
    metavar="ISO 639-1 LANGUAGE CODE",
    type=click.Choice(["all"] + list(get_all_languages().keys())),
    help="""The dataset languages to benchmark. Only relevant if `dataset` has not been
    specified. If "all" then the models will be benchmarked on all datasets. If not
    specified then this will use the `language` value.""",
)
@click.option(
    "--dataset-task",
    "-dt",
    default=None,
    show_default=True,
    multiple=True,
    type=click.Choice(list(get_all_dataset_tasks().keys())),
    help="The dataset tasks to consider.",
)
@click.option(
    "--batch-size",
    default="32",
    type=click.Choice(["1", "2", "4", "8", "16", "32"]),
    help="The batch size to use.",
)
@click.option(
    "--evaluate-train/--no-evaluate-train",
    default=False,
    show_default=True,
    help="Whether to evaluate the model on the training set.",
)
@click.option(
    "--progress-bar/--no-progress-bar",
    "-p/-np",
    default=True,
    show_default=True,
    help="Whether to show a progress bar.",
)
@click.option(
    "--raise-error-on-invalid-model/--no-raise-error-on-invalid-model",
    default=False,
    show_default=True,
    help="Whether to raise an error if a model is invalid.",
)
@click.option(
    "--verbose/--no-verbose",
    "-v/-nv",
    default=False,
    show_default=True,
    help="Whether extra input should be outputted during benchmarking",
)
@click.option(
    "--save-results/--no-save-results",
    "-s/-ns",
    default=True,
    show_default=True,
    help="Whether results should not be stored to disk.",
)
@click.option(
    "--cache-dir",
    "-c",
    default=".scandeval_cache",
    show_default=True,
    help="The directory where models are datasets are cached.",
)
@click.option(
    "--auth-token",
    type=str,
    default="",
    show_default=True,
    help="""The authentication token for the Hugging Face Hub. If specified then the
    `--use-auth-token` flag will be set to True.""",
)
@click.option(
    "--use-auth-token/--no-use-auth-token",
    default=False,
    show_default=True,
    help="""Whether an authentication token should be used, enabling evaluation of
    private models. Requires that you are logged in via the `huggingface-cli login`
    command.""",
)
@click.option(
    "--ignore-duplicates/--no-ignore-duplicates",
    default=True,
    show_default=True,
    help="""Whether to skip evaluation of models which have already been evaluated,
    with scores lying in the 'scandeval_benchmark_results.jsonl' file.""",
)
def benchmark(
    model_id: Tuple[str],
    dataset: Tuple[str],
    language: Tuple[str],
    model_language: Tuple[str],
    dataset_language: Tuple[str],
    raise_error_on_invalid_model: bool,
    dataset_task: Tuple[str],
    batch_size: str,
    evaluate_train: bool,
    progress_bar: bool,
    save_results: bool,
    cache_dir: str,
    use_auth_token: bool,
    auth_token: str,
    ignore_duplicates: bool,
    verbose: bool = False,
) -> None:
    """Benchmark pretrained language models on Scandinavian language tasks."""

    # Set up language variables
    model_ids = None if len(model_id) == 0 else list(model_id)
    datasets = None if len(dataset) == 0 else list(dataset)
    languages: List[str] = list(language)
    model_languages = None if len(model_language) == 0 else list(model_language)
    dataset_languages = None if len(dataset_language) == 0 else list(dataset_language)
    dataset_tasks = None if len(dataset_task) == 0 else list(dataset_task)
    batch_size_int = int(batch_size)
    auth: Union[str, bool] = auth_token if auth_token != "" else use_auth_token

    # Initialise the benchmarker class
    benchmarker = Benchmarker(
        language=languages,
        model_language=model_languages,
        dataset_language=dataset_languages,
        dataset_task=dataset_tasks,
        batch_size=batch_size_int,
        progress_bar=progress_bar,
        save_results=save_results,
        evaluate_train=evaluate_train,
        raise_error_on_invalid_model=raise_error_on_invalid_model,
        verbose=verbose,
        use_auth_token=auth,
        ignore_duplicates=ignore_duplicates,
        cache_dir=cache_dir,
    )

    # Perform the benchmark evaluation
    benchmarker(model_id=model_ids, dataset=datasets)
